#ifndef __TOOLS_FRAME_H
#define __TOOLS_FRAME_H

#include "list.h"

struct frame {
	struct list_head list;
	void *dev;

	unsigned long timestamp;
	uint8_t cb[64];
	union {
		struct list_head list;
		uint8_t raw_data[32];
	} dcb;

	unsigned int queue;
	unsigned int ref;
	unsigned int alloced;
	unsigned int total_len;
	unsigned int len;

	uint8_t *head;
	uint8_t *data;
	uint8_t *tail;

	uint8_t payload[];
};

static inline void *frame_put(struct frame *frame, unsigned int size)
{
	void *tmp;

	BUG_ON(frame->total_len + size > frame->alloced);

	frame->len += size;
	frame->total_len += size;

	tmp = (void *) frame->tail;
	frame->tail += size;

	BUG_ON(frame->tail > (frame->payload + frame->alloced));

	return tmp;
}

static inline void *frame_push(struct frame *frame, unsigned int size)
{
	frame->len += size;
	frame->data -= size;

	BUG_ON(frame->data < frame->payload);
	return frame->data;
}

static inline void frame_get(struct frame *frame)
{
	frame->ref++;
}

static inline void *frame_pull(struct frame *frame, unsigned int size)
{
	BUG_ON(frame->len < size);

	frame->len -= size;
	frame->total_len -= size;
	frame->data += size;

	return frame->data;
}

static inline void frame_reserve(struct frame *frame, unsigned int size)
{
	BUG_ON(frame->total_len + size > frame->alloced);
	BUG_ON(frame->len != 0);

	frame->total_len += size;
	frame->data += size;
	frame->tail += size;
}

static inline void frame_trim(struct frame *frame, unsigned int size)
{
	BUG_ON(size > frame->total_len);

	frame->len = size;
	frame->total_len = size;
	frame->data = frame->head;
	frame->tail = frame->head + size;
}

static inline struct frame *frame_alloc(unsigned int size)
{
	struct frame *tmp;

	tmp = malloc(size + sizeof(*tmp));
	if (tmp != NULL) {
		memset(tmp, 0, sizeof(*tmp));
		init_list_head(&tmp->list);
		init_list_head(&tmp->dcb.list);
		tmp->len = 0;
		tmp->total_len = 0;
		tmp->alloced = size;

		tmp->head = tmp->payload;
		tmp->data = tmp->payload;
		tmp->tail = tmp->payload;
		tmp->ref = 1;
	}
	return tmp;
}

static inline void frame_free(struct frame *frame)
{
	if (!--frame->ref)
		free(frame);
}

struct frame_queue {
	SDL_mutex *lock;
	size_t len;
	struct list_head list;
};

#define FRAME_WALK(pos, head)					\
	list_for_each_entry((pos), &(head)->list, list)

static inline void frame_queue_init(struct frame_queue *queue)
{
	queue->lock = SDL_CreateMutex();
	queue->len = 0;
	init_list_head(&queue->list);
}

static inline void frame_queue_kill(struct frame_queue *queue)
{
	SDL_DestroyMutex(queue->lock);
}

static inline unsigned int frame_queue_len(struct frame_queue *queue)
{
	return queue->len;
}

static inline bool frame_queue_empty(struct frame_queue *queue)
{
	return list_empty(&queue->list);
}

static inline void __frame_queue_head(struct frame_queue *queue, struct frame *frame)
{
	list_add_head(&frame->list, &queue->list);
	queue->len++;
}

static inline void frame_queue_head(struct frame_queue *queue, struct frame *frame)
{
	BUG_ON((SDL_mutexP(queue->lock) != 0));
	__frame_queue_head(queue, frame);
	SDL_mutexV(queue->lock);
}

static inline void __frame_queue_tail(struct frame_queue *queue, struct frame *frame)
{
	list_add_tail(&frame->list, &queue->list);
	queue->len++;
}

static inline void frame_queue_tail(struct frame_queue *queue, struct frame *frame)
{
	BUG_ON((SDL_mutexP(queue->lock) != 0));
	__frame_queue_tail(queue, frame);
	SDL_mutexV(queue->lock);
}

static inline void __frame_unlink(struct frame_queue *queue, struct frame *frame)
{
	list_del(&frame->list);
	queue->len--;
}

static inline struct frame *__frame_dequeue(struct frame_queue *queue)
{
	struct frame *tmp = NULL;

	if (!frame_queue_empty(queue)) {
		tmp = list_entry(queue->list.next, struct frame, list);
		__frame_unlink(queue, tmp);
	}

	return tmp;
}

static inline struct frame *frame_dequeue(struct frame_queue *queue)
{
	struct frame *tmp;

	BUG_ON((SDL_mutexP(queue->lock) != 0));

	tmp = __frame_dequeue(queue);
	SDL_mutexV(queue->lock);
	return tmp;
}

static inline void frame_queue_purge(struct frame_queue *queue)
{
	BUG_ON((SDL_mutexP(queue->lock) != 0));
	while (list_empty(&queue->list) == false)
		frame_free(__frame_dequeue(queue));

	SDL_mutexV(queue->lock);
}

#endif /* __TOOLS_FRAME_H */
